import textwrap

import pytest

from progtc.client import AsyncProgtcClient
from progtc.types import ExecuteCodeSuccess, MessageType


@pytest.fixture
def progtc_client(test_server: str) -> AsyncProgtcClient:
    return AsyncProgtcClient(
        base_url=test_server,
        api_key="1234567890",
    )


@pytest.mark.asyncio
async def test_client_ping(progtc_client: AsyncProgtcClient) -> None:
    result = await progtc_client.ping()
    assert result is True


@pytest.mark.asyncio
async def test_client_ping_unreachable() -> None:
    client = AsyncProgtcClient(
        base_url="http://127.0.0.1:1",
        api_key="test",
    )
    result = await client.ping()
    assert result is False


@pytest.mark.asyncio
async def test_client_wait_for_server_success(progtc_client: AsyncProgtcClient) -> None:
    await progtc_client.wait_for_server(timeout=5.0)


@pytest.mark.asyncio
async def test_client_wait_for_server_timeout() -> None:
    client = AsyncProgtcClient(
        base_url="http://127.0.0.1:1",
        api_key="test",
    )
    with pytest.raises(TimeoutError, match="Progtc server not ready after 0.1s"):
        await client.wait_for_server(timeout=0.1)


@pytest.mark.asyncio
async def test_client_execute_code_basic(progtc_client: AsyncProgtcClient) -> None:
    code = textwrap.dedent(
        """\
        print("Hello, world!")
        """,
    )
    result = await progtc_client.execute_code(code, tools={})

    assert isinstance(result, ExecuteCodeSuccess)
    assert result.stdout == "Hello, world!\n"


@pytest.mark.asyncio
async def test_client_execute_code_with_tool_call(
    progtc_client: AsyncProgtcClient,
) -> None:
    async def get_weather() -> str:
        return "Sunny"

    code = textwrap.dedent(
        """\
        from tools import get_weather
        print(await get_weather())
        """,
    )
    result = await progtc_client.execute_code(
        code,
        tools={"get_weather": get_weather},
    )

    assert isinstance(result, ExecuteCodeSuccess)
    assert result.stdout == "Sunny\n"


@pytest.mark.asyncio
async def test_client_execute_code_with_tool_call_args(
    progtc_client: AsyncProgtcClient,
) -> None:
    async def get_weather(city: str, country: str) -> str:
        return f"Weather in {city}, {country}: Rainy"

    code = textwrap.dedent(
        """\
        from tools import get_weather
        print(await get_weather("London", country="UK"))
        """,
    )
    result = await progtc_client.execute_code(
        code,
        tools={"get_weather": get_weather},
    )

    assert isinstance(result, ExecuteCodeSuccess)
    assert result.stdout == "Weather in London, UK: Rainy\n"


@pytest.mark.asyncio
async def test_client_execute_code_multiple_tool_calls(
    progtc_client: AsyncProgtcClient,
) -> None:
    async def increment(n: int) -> int:
        return n + 1

    code = textwrap.dedent(
        """\
        from tools import increment
        async def main():
            a = await increment(0)
            b = await increment(a)
            c = await increment(b)
            return [a, b, c]
        print(await main())
        """,
    )
    result = await progtc_client.execute_code(
        code,
        tools={"increment": increment},
    )

    assert isinstance(result, ExecuteCodeSuccess)
    assert result.stdout == "[1, 2, 3]\n"


@pytest.mark.asyncio
async def test_client_syntax_error(progtc_client: AsyncProgtcClient) -> None:
    code = "def main( invalid syntax"
    result = await progtc_client.execute_code(code, tools={})
    assert result.type == MessageType.SYNTAX_ERROR


@pytest.mark.asyncio
async def test_client_execution_error(progtc_client: AsyncProgtcClient) -> None:
    code = textwrap.dedent(
        """\
        async def main():
            raise ValueError("Something went wrong")
        print(await main())
        """,
    )
    result = await progtc_client.execute_code(code, tools={})
    assert result.type == MessageType.RUNTIME_ERROR
    assert "Something went wrong" in result.stderr



from typing import Annotated

import typer
import uvicorn
from rich import print as rprint

from progtc.server.config import server_config

app = typer.Typer(
    name="progtc",
    help="progtc CLI",
    add_completion=True,
)


@app.callback()
def callback() -> None:
    """Progtc - programmatic tool calling"""


ARTWORK = """
[bright_cyan]
    ██████╗ ██████╗  ██████╗  ██████╗ ████████╗ ██████╗
    ██╔══██╗██╔══██╗██╔═══██╗██╔════╝ ╚══██╔══╝██╔════╝
    ██████╔╝██████╔╝██║   ██║██║  ███╗   ██║   ██║     
    ██╔═══╝ ██╔══██╗██║   ██║██║   ██║   ██║   ██║     
    ██║     ██║  ██║╚██████╔╝╚██████╔╝   ██║   ╚██████╗
    ╚═╝     ╚═╝  ╚═╝ ╚═════╝  ╚═════╝    ╚═╝    ╚═════╝
                                               by capsa[/bright_cyan]
[dim]───────────────────────────────────────────────────────────[/dim]
    [bright_white]Come build agents with us: https://capsa.ai/careers[/bright_white]
[dim]───────────────────────────────────────────────────────────[/dim]
"""


@app.command()
def serve(
    host: Annotated[str, typer.Option(help="Host to bind to")] = "127.0.0.1",
    port: Annotated[int, typer.Option(help="Port to bind to")] = 8000,
    api_key: Annotated[
        str | None,
        typer.Option(
            help=(
                "API key for authentication (can also be set via "
                "PROGTC_API_KEY env var)"
            ),
            envvar="PROGTC_API_KEY",
        ),
    ] = None,
    tool_call_timeout: Annotated[
        float,
        typer.Option(help="Timeout for tool calls in seconds"),
    ] = 10.0,
    code_execution_timeout: Annotated[
        float,
        typer.Option(help="Timeout for code execution in seconds"),
    ] = 30.0,
) -> None:
    if api_key:
        server_config.api_key = api_key
    server_config.tool_call_timeout = tool_call_timeout
    server_config.code_execution_timeout = code_execution_timeout

    rprint(ARTWORK)

    uvicorn.run(
        "progtc.server.api:app",
        host=host,
        port=port,
    )


def main() -> None:
    app()


if __name__ == "__main__":
    main()



from pydantic import BaseModel


class ServerConfig(BaseModel):
    api_key: str = "1234567890"
    tool_call_timeout: float = 10.0
    code_execution_timeout: float = 30.0


server_config = ServerConfig()



from collections.abc import Generator
from contextlib import contextmanager

from progtc.server.code_execution import CodeExecutionProcess
from progtc.types import ToolCallResult


class CodeExecutionManager:
    """Manages the lifecycle of code execution processes."""

    def __init__(self) -> None:
        self._processes: dict[str, CodeExecutionProcess] = {}

    @contextmanager
    def run(
        self,
        code: str,
        tool_names: list[str],
    ) -> Generator[CodeExecutionProcess, None, None]:
        process = CodeExecutionProcess(code, tool_names)
        self._processes[process.execution_id] = process
        process.start()
        try:
            yield process
        finally:
            process.stop()
            del self._processes[process.execution_id]

    def send_tool_result(self, tool_call_result: ToolCallResult) -> None:
        self._processes[tool_call_result.execution_id].send_tool_result(
            tool_call_result,
        )



import asyncio
import time

from httpx import AsyncClient
from httpx_sse import aconnect_sse
from pydantic import TypeAdapter

from progtc.types import (
    ExecuteCodeError,
    ExecuteCodeMessage,
    ExecuteCodeSuccess,
    ToolCall,
    ToolHandler,
)

ExecuteCodeMessageAdapter = TypeAdapter[ExecuteCodeMessage](ExecuteCodeMessage)


class AsyncProgtcClient:
    def __init__(
        self,
        base_url: str,
        api_key: str,
        headers: dict[str, str] | None = None,
    ):
        self._base_url = base_url
        self._api_key = api_key
        self._headers = headers or {}

    async def ping(self, headers: dict[str, str] | None = None) -> bool:
        merged_headers = {**self._headers, **(headers or {})}
        async with AsyncClient(
            base_url=self._base_url, headers=merged_headers
        ) as client:
            try:
                response = await client.get("/ping")
                return response.status_code == 200
            except Exception:
                return False

    async def wait_for_server(
        self, timeout: float = 10.0, headers: dict[str, str] | None = None
    ) -> None:
        t0 = time.monotonic()
        while not await self.ping(headers=headers):
            await asyncio.sleep(0.5)
            if time.monotonic() - t0 > timeout:
                raise TimeoutError(f"Progtc server not ready after {timeout}s")

    async def execute_code(
        self,
        code: str,
        tools: dict[str, ToolHandler],
        headers: dict[str, str] | None = None,
    ) -> ExecuteCodeSuccess | ExecuteCodeError:
        merged_headers = {
            **self._headers,
            **(headers or {}),
            "X-Progtc-API-Key": self._api_key,
        }
        async with (
            AsyncClient(
                base_url=self._base_url,
                headers=merged_headers,
            ) as client,
            aconnect_sse(
                client,
                "POST",
                "/execute-code",
                json={
                    "tool_names": list(tools.keys()),
                    "code": code,
                },
            ) as event_source,
        ):
            async for event in event_source.aiter_sse():
                event_obj = ExecuteCodeMessageAdapter.validate_python(event.json())
                if isinstance(event_obj, ToolCall):
                    handler = tools[event_obj.tool_name]
                    result = await handler(*event_obj.args, **event_obj.kwargs)
                    await client.post(
                        "/tool-results",
                        json={
                            "execution_id": event_obj.execution_id,
                            "tool_call_id": event_obj.id,
                            "result": result,
                        },
                    )
                else:
                    return event_obj

        raise RuntimeError("Stream ended without result")



from collections.abc import Awaitable, Callable
from enum import Enum
from typing import Literal, TypeAlias
from uuid import uuid4

from pydantic import BaseModel, Field, JsonValue


class ExecuteCodeRequest(BaseModel):
    tool_names: list[str]
    code: str


class MessageType(str, Enum):
    TOOL_CALL = "tool_call"
    SUCCESS = "success"
    SYNTAX_ERROR = "syntax_error"
    RUNTIME_ERROR = "runtime_error"
    TIMEOUT_ERROR = "timeout_error"


class ToolCall(BaseModel):
    type: Literal[MessageType.TOOL_CALL] = MessageType.TOOL_CALL
    id: str = Field(default_factory=lambda: uuid4().hex)
    execution_id: str
    tool_name: str
    args: tuple[JsonValue, ...]
    kwargs: dict[str, JsonValue]


class ExecuteCodeSuccess(BaseModel):
    type: Literal[MessageType.SUCCESS] = MessageType.SUCCESS
    stdout: str
    stderr: str


class CodeSyntaxError(BaseModel):
    type: Literal[MessageType.SYNTAX_ERROR] = MessageType.SYNTAX_ERROR
    stdout: str
    stderr: str


class CodeRuntimeError(BaseModel):
    type: Literal[MessageType.RUNTIME_ERROR] = MessageType.RUNTIME_ERROR
    stdout: str
    stderr: str


class CodeTimeoutError(BaseModel):
    type: Literal[MessageType.TIMEOUT_ERROR] = MessageType.TIMEOUT_ERROR
    stdout: str
    stderr: str


ExecuteCodeError: TypeAlias = CodeSyntaxError | CodeRuntimeError | CodeTimeoutError


ExecuteCodeMessage: TypeAlias = ToolCall | ExecuteCodeSuccess | ExecuteCodeError


class ToolCallResult(BaseModel):
    execution_id: str
    tool_call_id: str
    result: JsonValue


ToolHandler: TypeAlias = Callable[..., Awaitable[JsonValue]]
